@@ -46,7 +46,7 @@ released under the GNU Lesser General Public License (LGPL).
 
 Website: U{http://www.lag.net/paramiko/}
 
-@version: 1.5 (paras)
+@version: 1.5.2 (rhydon)
 @author: Robey Pointer
 @contact: robey@lag.net
 @license: GNU Lesser General Public License (LGPL)
@@ -59,8 +59,9 @@ if sys.version_info < (2, 2):
 
 
 __author__ = "Robey Pointer <robey@lag.net>"
-__date__ = "02 Oct 2005"
-__version__ = "1.5 (paras)"
+__date__ = "04 Dec 2005"
+__version__ = "1.5.2 (rhydon)"
+__version_info__ = (1, 5, 2)
 __license__ = "GNU Lesser General Public License (LGPL)"
 
 
@@ -116,6 +117,7 @@ __all__ = [ 'Transport',
             'PasswordRequiredException',
             'BadAuthenticationType',
             'SFTP',
+            'SFTPFile',
             'SFTPHandle',
             'SFTPClient',
             'SFTPServer',
