@@ -20,10 +20,13 @@
 Client-mode SFTP support.
 """
 
+import errno
 import os
-from sftp import *
-from sftp_attr import SFTPAttributes
-from sftp_file import SFTPFile
+import threading
+import weakref
+from paramiko.sftp import *
+from paramiko.sftp_attr import SFTPAttributes
+from paramiko.sftp_file import SFTPFile
 
 
 def _to_unicode(s):
@@ -55,8 +58,11 @@ class SFTPClient (BaseSFTP):
         self.sock = sock
         self.ultra_debug = False
         self.request_number = 1
+        # lock for request_number
+        self._lock = threading.Lock()
         self._cwd = None
-        self._expecting = []
+        # request # -> SFTPFile
+        self._expecting = weakref.WeakValueDictionary()
         if type(sock) is Channel:
             # override default logger
             transport = self.sock.get_transport()
@@ -147,10 +153,10 @@ class SFTPClient (BaseSFTP):
         self._request(CMD_CLOSE, handle)
         return filelist
 
-    def open(self, filename, mode='r', bufsize=-1):
+    def file(self, filename, mode='r', bufsize=-1):
         """
         Open a file on the remote server.  The arguments are the same as for
-        python's built-in C{open} (aka C{file}).  A file-like object is
+        python's built-in C{file} (aka C{open}).  A file-like object is
         returned, which closely mimics the behavior of a normal python file
         object.
 
@@ -160,6 +166,11 @@ class SFTPClient (BaseSFTP):
         existing file), C{'a+'} for reading/appending.  The python C{'b'} flag
         is ignored, since SSH treats all files as binary.  The C{'U'} flag is
         supported in a compatible way.
+        
+        Since 1.5.2, an C{'x'} flag indicates that the operation should only
+        succeed if the file was created and did not previously exist.  This has
+        no direct mapping to python's file flags, but is commonly known as the
+        C{O_EXCL} flag in posix.
 
         The file will be buffered in standard python style by default, but
         can be altered with the C{bufsize} parameter.  C{0} turns off
@@ -181,12 +192,14 @@ class SFTPClient (BaseSFTP):
         imode = 0
         if ('r' in mode) or ('+' in mode):
             imode |= SFTP_FLAG_READ
-        if ('w' in mode) or ('+' in mode):
+        if ('w' in mode) or ('+' in mode) or ('a' in mode):
             imode |= SFTP_FLAG_WRITE
         if ('w' in mode):
             imode |= SFTP_FLAG_CREATE | SFTP_FLAG_TRUNC
         if ('a' in mode):
-            imode |= SFTP_FLAG_APPEND | SFTP_FLAG_CREATE | SFTP_FLAG_WRITE
+            imode |= SFTP_FLAG_CREATE | SFTP_FLAG_APPEND
+        if ('x' in mode):
+            imode |= SFTP_FLAG_CREATE | SFTP_FLAG_EXCL
         attrblock = SFTPAttributes()
         t, msg = self._request(CMD_OPEN, filename, imode, attrblock)
         if t != CMD_HANDLE:
@@ -196,7 +209,7 @@ class SFTPClient (BaseSFTP):
 
     # python has migrated toward file() instead of open().
     # and really, that's more easily identifiable.
-    file = open
+    open = file
 
     def remove(self, path):
         """
@@ -489,6 +502,7 @@ class SFTPClient (BaseSFTP):
         @since: 1.4
         """
         fr = self.file(remotepath, 'rb')
+        fr.prefetch()
         fl = file(localpath, 'wb')
         size = 0
         while True:
@@ -508,27 +522,32 @@ class SFTPClient (BaseSFTP):
 
 
     def _request(self, t, *arg):
-        num = self._async_request(t, *arg)
+        num = self._async_request(type(None), t, *arg)
         return self._read_response(num)
     
-    def _async_request(self, t, *arg):
-        msg = Message()
-        msg.add_int(self.request_number)
-        for item in arg:
-            if type(item) is int:
-                msg.add_int(item)
-            elif type(item) is long:
-                msg.add_int64(item)
-            elif type(item) is str:
-                msg.add_string(item)
-            elif type(item) is SFTPAttributes:
-                item._pack(msg)
-            else:
-                raise Exception('unknown type for %r type %r' % (item, type(item)))
-        self._send_packet(t, str(msg))
-        num = self.request_number
-        self.request_number += 1
-        self._expecting.append(num)
+    def _async_request(self, fileobj, t, *arg):
+        # this method may be called from other threads (prefetch)
+        self._lock.acquire()
+        try:
+            msg = Message()
+            msg.add_int(self.request_number)
+            for item in arg:
+                if type(item) is int:
+                    msg.add_int(item)
+                elif type(item) is long:
+                    msg.add_int64(item)
+                elif type(item) is str:
+                    msg.add_string(item)
+                elif type(item) is SFTPAttributes:
+                    item._pack(msg)
+                else:
+                    raise Exception('unknown type for %r type %r' % (item, type(item)))
+            num = self.request_number
+            self._expecting[num] = fileobj
+            self._send_packet(t, str(msg))
+            self.request_number += 1
+        finally:
+            self._lock.release()
         return num
 
     def _read_response(self, waitfor=None):
@@ -537,18 +556,29 @@ class SFTPClient (BaseSFTP):
             msg = Message(data)
             num = msg.get_int()
             if num not in self._expecting:
-                raise SFTPError('Expected response from %r, got response #%d' %
-                    (self._expected, num))
-            self._expecting.remove(num)
-            if t == CMD_STATUS:
-                self._convert_status(msg)
-            if (waitfor is None) or (num == waitfor):
-                break
-        return t, msg
+                # might be response for a file that was closed before responses came back
+                self._log(DEBUG, 'Unexpected response #%d' % (num,))
+                if waitfor is None:
+                    # just doing a single check
+                    return
+                continue
+            fileobj = self._expecting[num]
+            del self._expecting[num]
+            if num == waitfor:
+                # synchronous
+                if t == CMD_STATUS:
+                    self._convert_status(msg)
+                return t, msg
+            if fileobj is not type(None):
+                fileobj._async_response(t, msg)
+            if waitfor is None:
+                # just doing a single check
+                return
 
-    def _finish_responses(self):
-        while len(self._expecting) > 0:
+    def _finish_responses(self, fileobj):
+        while fileobj in self._expecting.values():
             self._read_response()
+            fileobj._check_exception()
 
     def _convert_status(self, msg):
         """
@@ -560,6 +590,11 @@ class SFTPClient (BaseSFTP):
             return
         elif code == SFTP_EOF:
             raise EOFError(text)
+        elif code == SFTP_NO_SUCH_FILE:
+            # clever idea from john a. meinel: map the error codes to errno
+            raise IOError(errno.ENOENT, text)
+        elif code == SFTP_PERMISSION_DENIED:
+            raise IOError(errno.EACCES, text)
         else:
             raise IOError(text)
     
