@@ -23,11 +23,15 @@ a real actual sftp server is contacted, and a new folder is created there to
 do test file operations in (so no existing files will be harmed).
 """
 
-import sys, os
+import logging
+import os
 import random
-import logging, threading
+import sys
+import threading
+import time
+import unittest
 
-import paramiko, unittest
+import paramiko
 from stub_sftp import StubServer, StubSFTPServer
 from loop import LoopSocket
 
@@ -58,7 +62,7 @@ liver insulin receptors. Their sensitivity to insulin is, however, similarly
 decreased compared with chicken.
 '''
 
-FOLDER = os.environ.get('TEST_FOLDER', 'temp-testing')
+FOLDER = os.environ.get('TEST_FOLDER', 'temp-testing000')
 
 sftp = None
 tc = None
@@ -117,7 +121,6 @@ class SFTPTest (unittest.TestCase):
         ts.start_server(event, server)
         tc.connect(username='slowdive', password='pygmalion')
         event.wait(1.0)
-#        self.assert_(self.ts.is_active())
 
         sftp = paramiko.SFTP.from_transport(tc)
     init_loopback = staticmethod(init_loopback)
@@ -128,7 +131,14 @@ class SFTPTest (unittest.TestCase):
     set_big_file_test = staticmethod(set_big_file_test)
         
     def setUp(self):
-        sftp.mkdir(FOLDER)
+        global FOLDER
+        for i in xrange(1000):
+            FOLDER = FOLDER[:-3] + '%03d' % i
+            try:
+                sftp.mkdir(FOLDER)
+                break
+            except (IOError, OSError):
+                pass
 
     def tearDown(self):
         sftp.rmdir(FOLDER)
@@ -432,12 +442,13 @@ class SFTPTest (unittest.TestCase):
 
     def test_E_big_file(self):
         """
-        write a 1MB file, with no linefeeds, using line buffering.
+        write a 1MB file with no buffering.
         """
         global g_big_file_test
         if not g_big_file_test:
             return
         kblob = (1024 * 'x')
+        start = time.time()
         try:
             f = sftp.open('%s/hongry.txt' % FOLDER, 'w')
             for n in range(1024):
@@ -448,6 +459,18 @@ class SFTPTest (unittest.TestCase):
             sys.stderr.write(' ')
 
             self.assertEqual(sftp.stat('%s/hongry.txt' % FOLDER).st_size, 1024 * 1024)
+            end = time.time()
+            sys.stderr.write('%ds ' % round(end - start))
+            
+            start = time.time()
+            f = sftp.open('%s/hongry.txt' % FOLDER, 'r')
+            for n in range(1024):
+                data = f.read(1024)
+                self.assertEqual(data, kblob)
+            f.close()
+
+            end = time.time()
+            sys.stderr.write('%ds ' % round(end - start))
         finally:
             sftp.remove('%s/hongry.txt' % FOLDER)
 
@@ -459,6 +482,43 @@ class SFTPTest (unittest.TestCase):
         if not g_big_file_test:
             return
         kblob = (1024 * 'x')
+        start = time.time()
+        try:
+            f = sftp.open('%s/hongry.txt' % FOLDER, 'w')
+            f.set_pipelined(True)
+            for n in range(1024):
+                f.write(kblob)
+                if n % 128 == 0:
+                    sys.stderr.write('.')
+            f.close()
+            sys.stderr.write(' ')
+
+            self.assertEqual(sftp.stat('%s/hongry.txt' % FOLDER).st_size, 1024 * 1024)
+            end = time.time()
+            sys.stderr.write('%ds ' % round(end - start))
+            
+            start = time.time()
+            f = sftp.open('%s/hongry.txt' % FOLDER, 'r')
+            f.prefetch()
+            for n in range(1024):
+                data = f.read(1024)
+                self.assertEqual(data, kblob)
+            f.close()
+
+            end = time.time()
+            sys.stderr.write('%ds ' % round(end - start))
+        finally:
+            sftp.remove('%s/hongry.txt' % FOLDER)
+
+    def test_G_lots_of_prefetching(self):
+        """
+        prefetch a 1MB file a bunch of times, discarding the file object
+        without using it, to verify that paramiko doesn't get confused.
+        """
+        global g_big_file_test
+        if not g_big_file_test:
+            return
+        kblob = (1024 * 'x')
         try:
             f = sftp.open('%s/hongry.txt' % FOLDER, 'w')
             f.set_pipelined(True)
@@ -470,27 +530,62 @@ class SFTPTest (unittest.TestCase):
             sys.stderr.write(' ')
 
             self.assertEqual(sftp.stat('%s/hongry.txt' % FOLDER).st_size, 1024 * 1024)
+
+            for i in range(10):
+                f = sftp.open('%s/hongry.txt' % FOLDER, 'r')
+                f.prefetch()
+            f = sftp.open('%s/hongry.txt' % FOLDER, 'r')
+            f.prefetch()
+            for n in range(1024):
+                data = f.read(1024)
+                self.assertEqual(data, kblob)
+                if n % 128 == 0:
+                    sys.stderr.write('.')
+            f.close()
+            sys.stderr.write(' ')
         finally:
             sftp.remove('%s/hongry.txt' % FOLDER)
         
-    def test_G_big_file_big_buffer(self):
+    def test_H_big_file_big_buffer(self):
         """
         write a 1MB file, with no linefeeds, and a big buffer.
         """
         global g_big_file_test
         if not g_big_file_test:
             return
-        kblob = (1024 * 1024 * 'x')
+        mblob = (1024 * 1024 * 'x')
         try:
             f = sftp.open('%s/hongry.txt' % FOLDER, 'w', 128 * 1024)
-            f.write(kblob)
+            f.write(mblob)
             f.close()
 
             self.assertEqual(sftp.stat('%s/hongry.txt' % FOLDER).st_size, 1024 * 1024)
         finally:
             sftp.remove('%s/hongry.txt' % FOLDER)
+    
+    def test_I_big_file_renegotiate(self):
+        """
+        write a 1MB file, forcing key renegotiation in the middle.
+        """
+        global g_big_file_test
+        if not g_big_file_test:
+            return
+        t = sftp.sock.get_transport()
+        t.packetizer.REKEY_BYTES = 512 * 1024
+        k32blob = (32 * 1024 * 'x')
+        try:
+            f = sftp.open('%s/hongry.txt' % FOLDER, 'w', 128 * 1024)
+            for i in xrange(32):
+                f.write(k32blob)
+            f.close()
             
-    def test_H_realpath(self):
+            self.assertEqual(sftp.stat('%s/hongry.txt' % FOLDER).st_size, 1024 * 1024)
+            self.assertNotEquals(t.H, t.session_id)
+        finally:
+            sftp.remove('%s/hongry.txt' % FOLDER)
+            t.packetizer.REKEY_BYTES = pow(2, 30)
+
+    def test_J_realpath(self):
         """
         test that realpath is returning something non-empty and not an
         error.
@@ -501,7 +596,7 @@ class SFTPTest (unittest.TestCase):
         self.assert_(len(f) > 0)
         self.assertEquals(os.path.join(pwd, FOLDER), f)
 
-    def test_I_mkdir(self):
+    def test_K_mkdir(self):
         """
         verify that mkdir/rmdir work.
         """
@@ -524,7 +619,7 @@ class SFTPTest (unittest.TestCase):
         except IOError:
             pass
     
-    def test_J_chdir(self):
+    def test_L_chdir(self):
         """
         verify that chdir/getcwd work.
         """
@@ -561,7 +656,7 @@ class SFTPTest (unittest.TestCase):
             except:
                 pass
 
-    def test_K_get_put(self):
+    def test_M_get_put(self):
         """
         verify that get/put work.
         """
@@ -590,7 +685,7 @@ class SFTPTest (unittest.TestCase):
         os.unlink(localname)
         sftp.unlink(FOLDER + '/bunny.txt')
 
-    def test_L_check(self):
+    def test_N_check(self):
         """
         verify that file.check() works against our own server.
         (it's an sftp extension that we support, and may be the only ones who
@@ -611,3 +706,35 @@ class SFTPTest (unittest.TestCase):
                               paramiko.util.hexify(sum))
         finally:
             sftp.unlink(FOLDER + '/kitty.txt')
+
+    def test_O_x_flag(self):
+        """
+        verify that the 'x' flag works when opening a file.
+        """
+        f = sftp.open(FOLDER + '/unusual.txt', 'wx')
+        f.close()
+        
+        try:
+            try:
+                f = sftp.open(FOLDER + '/unusual.txt', 'wx')
+                self.fail('expected exception')
+            except IOError, x:
+                pass
+        finally:
+            sftp.unlink(FOLDER + '/unusual.txt')
+    
+    def test_P_utf8(self):
+        """
+        verify that unicode strings are encoded into utf8 correctly.
+        """
+        f = sftp.open(FOLDER + '/something', 'w')
+        f.write('okay')
+        f.close()
+        
+        try:
+            sftp.rename(FOLDER + '/something', FOLDER + u'/\u00fcnic\u00f8de')
+            sftp.open(FOLDER + '/\xc3\xbcnic\xc3\xb8\x64\x65', 'r')
+        except Exception, e:
+            self.fail('exception ' + e)
+        sftp.unlink(FOLDER + '/\xc3\xbcnic\xc3\xb8\x64\x65')
+
